Option _Explicit
'$INCLUDE:'./gx/gx.bi'
_Title "Legend of GX!"

Const SCENE_WIDTH = 256
Const SCENE_HEIGHT = 240
Const SCENE_YOFFSET = 64
Const WALK_SPEED = 80
Const ANIMATE_SPEED = 10
Const RIGHT = 1
Const LEFT = 2
Const DOWN = 3
Const UP = 4
Const False = 0
Const True = Not False

Dim Shared toggleDebug As Integer
Dim Shared mode As Integer
Dim Shared link As Long
Dim Shared status As Long
Dim Shared music As Long
Dim Shared mapx As Integer: mapx = 7
Dim Shared mapy As Integer: mapy = 7
Dim Shared pagemove
Dim Shared pagex: pagex = -1
Dim Shared pagey: pagey = -1
Dim Shared gameFont As Integer
Dim Shared roomMode As Integer
Dim Shared rubies As Integer
Dim Shared keys As Integer
Dim Shared bombs As Integer
Dim Shared opos As GXPosition
Dim Shared omap As GXPosition
Dim Shared moveHistory As String
Dim Shared lostWoods As Integer
opos.x = 1911
opos.y = 1310

GXSceneCreate SCENE_WIDTH, SCENE_HEIGHT
GXSceneStart
System

Sub GXOnGameEvent (e As GXEvent)
    Select Case e.event
        Case GXEVENT_INIT: OnInit
        Case GXEVENT_UPDATE: OnUpdate
        Case GXEVENT_DRAWSCREEN: DrawHUD
        Case GXEVENT_COLLISION_TILE: OnCollisionTile e
    End Select
End Sub

Sub OnInit
    'GXSoundMuted True
    'GXFullScreen True
    If Not GXFullScreen Then GXSceneScale 2
    gameFont = GXFontCreate("img/z-font.png", 8, 8, "02468ACEGIKMOQSU" + GX_CRLF + _
                                                    "13579BDFHJLNPRTV" + GX_CRLF + _
                                                    "WX  ,'.?" + gx_CRLF + _
                                                    "XZ  !&" + chr$(34) + "-")
    music = GXSoundLoad("snd/track3.mp3")
    link = GXEntityCreate("img/link.png", 16, 16, 2)

    If status = 0 Then status = GXScreenEntityCreate("img/status.png", 256, SCENE_YOFFSET, 1)

    LoadOverworld
End Sub


Sub OnUpdate
    ' If Esc is pressed, exit the game
    If GXKeyDown(GXKEY_ESC) Then
        GXSceneStop
    End If

    ' Toggle debug mode when F1 key is pressed
    If GXKeyDown(GXKEY_F1) Then toggleDebug = GX_TRUE
    If Not GXKeyDown(GXKEY_F1) And toggleDebug Then
        GXDebug Not GXDebug
        toggleDebug = GX_FALSE
    End If

    HandlePlayerControls

    If roomMode Then
        If GXEntityY(link) > GXMapRows * GXTilesetHeight Then
            ReturnToOverworld
        End If
    Else
        If pagemove = 0 Then
            CheckPageMove link
        Else
            PageSwipe
        End If
    End If
End Sub

Sub OnCollisionTile (e As GXEvent)
    ' Simple test to see if the tile is one we can walk on
    ' Could be replaced with a more sophisiticated collision layer
    Dim tile As Integer
    tile = GXMapTile(e.collisionTileX, e.collisionTileY, 1)
    IF tile <> 385 And tile <> 382 And tile <> 555 And tile <> 560 And _
       tile <> 522 And tile <> 519 And tile <> 522 And _
       tile <> 348 And tile <> 380 And tile <> 65  And _
       tile <> 11  And tile <> 43  And tile <> 1   And _
       tile <> 403 And tile <> 404 And tile <> 435 And _
       tile <> 436 And tile <> 347 And tile <> 379 And tile <> 0 THEN
        e.collisionResult = 1
    End If

    ' Test for map triggers
    tile = GXMapTile(e.collisionTileX, e.collisionTileY, 3)
    If tile <> 0 Then
        LoadCave
    End If
End Sub

Sub HandlePlayerControls
    If GXKeyDown(GXKEY_DOWN) Then
        GXEntityVX link, 0
        GXEntityVY link, WALK_SPEED
        GXEntityAnimate link, DOWN, ANIMATE_SPEED

    ElseIf GXKeyDown(GXKEY_UP) Then
        GXEntityVX link, 0
        GXEntityVY link, -WALK_SPEED
        GXEntityAnimate link, UP, ANIMATE_SPEED

    ElseIf GXKeyDown(GXKEY_RIGHT) Then
        GXEntityVX link, WALK_SPEED
        GXEntityVY link, 0
        GXEntityAnimate link, RIGHT, ANIMATE_SPEED

    ElseIf GXKeyDown(GXKEY_LEFT) Then
        GXEntityVX link, -WALK_SPEED
        GXEntityVY link, 0
        GXEntityAnimate link, LEFT, ANIMATE_SPEED

    Else
        GXEntityVX link, 0
        GXEntityVY link, 0
        GXEntityAnimateStop link
    End If

    If pagemove <> 0 Then
        ' don't allow movement while page swipe in progress
        GXEntityVX link, 0
        GXEntityVY link, 0
    End If
End Sub

Sub LoadOverworld
    GXMapLoad "map/overworld.gxm"
    GXMapLayerVisible 2, False ' Hide the maze layer
    GXMapLayerVisible 3, False ' Hide the collision/trigger layer


    GXEntityPos link, opos.x, opos.y
    GXScenePos mapx * GXSceneWidth, mapy * (GXSceneHeight - SCENE_YOFFSET) - SCENE_YOFFSET
    GXEntityFrameSet link, 3, 1

    GXEntityCollisionOffset link, 2, 6, 2, 1
    GXSoundRepeat music
End Sub

Sub LoadCave
    ' remember the overworld position
    opos.x = GXEntityX(link)
    opos.y = GXEntityY(link)
    omap.x = mapx
    omap.y = mapy
    GXSoundPause music
    GXMapLoad "map/cave.gxm"
    GXScenePos 0, -SCENE_YOFFSET
    GXEntityPos link, 120, 160
    mapx = 1
    mapy = 1
    roomMode = True
End Sub

Sub ReturnToOverworld
    roomMode = False
    mapx = omap.x
    mapy = omap.y
    opos.y = opos.y + 8
    GXEntityFrameSet link, DOWN, 1
    LoadOverworld
End Sub

Sub DrawHUD
    ' draw map position
    Dim x As Integer
    Dim y As Integer
    x = mapx * 4 + 16
    y = mapy * 4 + 24
    Line (x, y)-(x + 2, y + 2), _RGB(134, 218, 0), BF

    GXDrawText gameFont, 104, 24, _Trim$(Str$(rubies))
    GXDrawText gameFont, 104, 41, _Trim$(Str$(keys))
    GXDrawText gameFont, 104, 49, _Trim$(Str$(bombs))

    ' debugging for lost woods
    If GXDebug Then
        GXDrawText GXFONT_DEFAULT, 1, 1, moveHistory
        If lostWoods Then
            GXDrawText GXFONT_DEFAULT, 1, 9, "Lost Woods"
        End If
    End If
End Sub

Sub CheckPageMove (entity)
    If (GXEntityX(entity) + GXEntityWidth(entity)) < mapx * GXSceneWidth Then
        pagemove = -5
        pagex = (mapx - 1) * GXSceneWidth
        mapx = mapx - 1
        LastMove "W"

    ElseIf GXEntityX(entity) > (mapx + 1) * GXSceneWidth Then
        pagemove = 5
        pagex = (mapx + 1) * GXSceneWidth
        mapx = mapx + 1
        LastMove "E"

    ElseIf (GXEntityY(entity) + GXEntityHeight(entity)) < mapy * (GXSceneHeight - SCENE_YOFFSET) Then
        pagemove = -5
        pagey = (mapy - 1) * (GXSceneHeight - SCENE_YOFFSET) - SCENE_YOFFSET
        mapy = mapy - 1
        LastMove "N"

    ElseIf GXEntityY(entity) > (mapy + 1) * (GXSceneHeight - SCENE_YOFFSET) Then
        pagemove = 5
        pagey = (mapy + 1) * (GXSceneHeight - SCENE_YOFFSET) - SCENE_YOFFSET
        mapy = mapy + 1
        LastMove "S"
    End If
End Sub

Sub LastMove (direction As String)
    ' remember the last four moves for maze
    moveHistory = Right$(moveHistory + direction, 4)

    If lostWoods Then
        If direction = "E" Or moveHistory = "NWSW" Then
            lostWoods = False
            GXMapLayerVisible 2, False
        End If
    End If
End Sub

Function LastMove$
    LastMove = Right$(moveHistory, 1)
End Function

Sub PageSwipe
    If pagex <> -1 Then
        If (pagemove < 1 And GXSceneX > pagex) Or (pagemove > 1 And GXSceneX < pagex) Then
            GXSceneMove pagemove, 0
        Else
            GXScenePos pagex, GXSceneY
            pagemove = 0
            pagex = -1
            EndPageSwipe
        End If
    End If

    If pagey <> -1 Then
        If (pagemove < 1 And GXSceneY > pagey) Or (pagemove > 1 And GXSceneY < pagey) Then
            GXSceneMove 0, pagemove
        Else
            GXScenePos GXSceneX, pagey
            pagemove = 0
            pagey = -1
            EndPageSwipe
        End If
    End If
End Sub

Sub EndPageSwipe
    If lostWoods Then
        If LastMove = "N" Then
            GXScenePos GXSceneX, GXSceneY + (GXSceneHeight - SCENE_YOFFSET)
            GXEntityPos link, GXEntityX(link), GXEntityY(link) + (GXSceneHeight - SCENE_YOFFSET)
            mapy = mapy + 1
        ElseIf LastMove = "S" Then
            GXScenePos GXSceneX, GXSceneY - (GXSceneHeight - SCENE_YOFFSET)
            GXEntityPos link, GXEntityX(link), GXEntityY(link) - (GXSceneHeight - SCENE_YOFFSET)
            mapy = mapy - 1
        ElseIf LastMove = "W" Then
            GXScenePos GXSceneX + GXSceneWidth, GXSceneY
            GXEntityPos link, GXEntityX(link) + GXSceneWidth, GXEntityY(link)
            mapx = mapx + 1
        End If
    End If

    If mapx = 1 And mapy = 6 Then
        lostWoods = True
        GXMapLayerVisible 2, True
    End If

End Sub

'$INCLUDE:'./gx/gx.bm'

